/** 									// -o-
 * Slider.pde
 *
 * Slider methods complement those of Trigger.  Slider building
 * blocks include:
 * 
 *   . value(), valuePerPosition(), valueRange()
 *   . knobPosition()
 *   . knobDimentions() 
 *   . knobLag()  
 *   . triggerAction()
 *
 * The SliderAction class defines what happens when a Slider is triggered.
 * The default action simply sends value in an OSC message to oscPath.
 * Customize this action by extending the class, then install it with
 * Slider.triggerAction().
 *
 * To paint the Slider value as text over the knob, use the
 * following code in displayTriggers():
 *
 *     slidObj.writing( slidObj.value() );
 *     slidObj.display();
 *
 * Slider delay and lag are orthogonal.
 *
 * Whenever possible, knob position calculations are derived from value.  
 * value is rounded for display purposes only.  value is also rounded
 * before placing into an OSC message. 
 *
 * ----------------------------------------------------------------------
 * Copyright © 2011, David Reeder.  All rights reserved.
 * Please see the LICENSE file in this distribution for license details.
 * ----------------------------------------------------------------------
 */


// ------------------------------------------------ -o--
class Slider extends Trigger
{
  String  className = "Slider";

  int  centerX, centerY; 	// absolute center
  int  wl, wr, hh, hl; 		// geographically: width left/right, 
  				//   height high/low

  int  length, 			// length of slider
       tailLength; 		// width of slider tails
  int  lengthAxis; 		// axis parallel to length

  public static final int  AXIS_HORIZONTAL = 1;
  public static final int  AXIS_VERTICAL = 2;


  int  	   knobWidth; 		// parallel to lengthAxis
  int      knobHeight;     	// perpendicular to lengthAxis
  int      knob_position; 	// swings along lengthAxis
  int  	   knob_lag;            // ms to move from current position to
  				//   center of difference
  boolean  knobLagActive = false;
  				// transitionining between points
				//   triggered by differences

  float  valueRangeLow, valueRangeHigh; 	
  				// range of values represented by slider
  float  value = 0;		// current value of the slider 

  float  valueSequence, 	// value along step sequence 
         valueIncrement; 	// value increment per knob lag step
  int    valueSteps; 		// number of steps in knob lag sequence


  int  knobTextHOffset = -15; 	// general height adjustment for knob text

  SliderAction actionObj; 	// defines trigger action that is
  				//   activated when value is updated



  // ------------------------- -o-
  // constructors
  //
  Slider() { /*EMPTY*/ } 

  Slider( String path_, String color_, 
	  int x_, int y_, int len_, int taillen_, int axis_ )
  {
    super(path_, color_);
    setSlid(x_, y_, len_, taillen_, axis_);
  }

  private void setSlid(int x_, int y_, int len_, int taillen_, int axis_)
  {
    centerX  	= x_;
    centerY 	= y_;

    length 	= len_;
    tailLength 	= taillen_;
    lengthAxis 	= axis_;

    checkBounds();

    pixelDiffCountMinimum = tailLength / _pixelDiffCountRatio;


    // DEFAULTS  XXX
    //
    actionObj = new SliderAction(this);
    oscMsg(-1); 		// XXX  placeholder

    knobDimentions(30, 20);
    colorFillSolid = COLOR_SOLID;

    valueRange(0, 100);
    knobPosition(50);

    knobLag(0);

  }  // setSlid()


  protected void checkBounds() 						// -o-
  {
    int  lengthMin 	= 20; 	// XXX  (default)
    int  tailLengthMin 	= 10; 	// XXX  (default)


    if (! ((AXIS_HORIZONTAL == lengthAxis) || (AXIS_VERTICAL == lengthAxis)) ) 
    {
      ERRORLN( className 
                 + "(): unknown axis type (lengthAxis=" + lengthAxis +")" );
    }

    if ((length < lengthMin) || (tailLength < tailLengthMin))
    {
      ERRORLN( className + "() is too small -- "
		 +"length="+ length +"; tailLength=" + tailLength );
    }

    
    if (AXIS_HORIZONTAL == lengthAxis) {
      wl = centerX - (length/2) - (length%2);
      wr = centerX + (length/2);

      hh = centerY - (tailLength/2) - (tailLength%2);
      hl = centerY + (tailLength/2);

    } else { 	// AXIS_VERTICAL
      hh = centerY - (length/2) - (length%2);
      hl = centerY + (length/2);

      wl = centerX - (tailLength/2) - (tailLength%2);
      wr = centerX + (tailLength/2);
    }

    if (! ((wl>=0) && (wr<width) && (hh>=0) && (hl<height)) )
    {
      ERRORLN( className + "() is out of bounds -- "
		 + " x=" + wl + ".." + wr + "; y=" + hh + ".." + hl );
    }
  }  // checkBounds()



  // ------------------------- -o-
  // helpers
  //
  String toString()  { return toString(""); }

  String toString(String prefix) 
  {
    String  class_desc = className + "(" + osc_path +", "+ colorName +")";


    String  			 axis = "AXIS_HORIZONTAL";
    if (AXIS_VERTICAL == lengthAxis)  { axis = "AXIS_VERTICAL"; }

    String  class_info1 = "xy=("+ centerX +","+ centerY +")"
    			    +" len/tail=" + length +"/"+ tailLength
			    +" ("+ axis +")";

    String  class_info2 = "KNOB  pos="+ knob_position +" lag="+ knob_lag +
			    printToggle("", (true==knobLagActive)?1:0, " ")
			  +"   VALUE  "+ value +" ["+ 
			    valueRangeLow +","+ valueRangeHigh +"] inc="+
			    round100(valueIncrement) +" steps="+ valueSteps ;


    if ( (prefix != null) && (0 < prefix.length()) ) 
    {
      class_desc = prefix +" = "+ class_desc;
    } 

    return super.toString(class_desc, class_info1, class_info2);
  }  // toString()


  Slider display()							
  {
    // draw slider
    //
    super.display();

    if (AXIS_HORIZONTAL == lengthAxis) {
      line(wl, centerY, wr, centerY);
      line(wl, hh, wl, hl);
      line(wr, hh, wr, hl);

    } else { 	// AXIS_VERTICAL 
      line(centerX, hh, centerX, hl);
      line(wl, hh, wr, hh);
      line(wl, hl, wr, hl);
    }


    // draw knob and text
    //
    super.display();
    rectMode(CENTER);

    if (AXIS_HORIZONTAL == lengthAxis) {
      rect(knob_position, centerY, knobWidth, knobHeight);
    } else {
      rect(centerX, knob_position, knobHeight, knobWidth);
    }

    rectMode(CORNER); 


    if (true == text_enabled) { 
      fill(colorVal, COLOR_VISIBLE);

      // XXX  text adjustment per knob, per slider orientation 
      //
      if (AXIS_HORIZONTAL == lengthAxis) {
	text(text_string, knob_position, centerY +knobTextHOffset);
      } else {
	text(text_string, centerX, hh +knobTextHOffset+7);
      }
    }

    return this;
  }  // display()



  // ------------------------- -o-
  // methods
  //
  //   Slider 		valueRange(float rangelow_, float rangehigh_)
  //   float 		value()  
  //   Slider 		value(float val_)
  //   private float 	valuePerPosition(int newpos_)
  //
  //   Slider 		knobPosition(int percentage_) 	
  //   private Slider 	resetKnobPosition() 
  //   Slider 		knobDimentions(int w_, int h_) 
  //
  //   int 		knobLag()
  //   Slider 		knobLag(int lag_)  
  //   private boolean 	lagStep() 
  //   private boolean 	lagStep(int newpos_)
  //
  //   void  triggerAction(SliderAction slidaction_) 
  //   void  triggerAction() 
  //
  //   boolean      isActive() 	
  //   private int  scanSlider()  	
  //

  // valueRange()
  // 
  // reset range of values reported by slider;
  // translate existing value into new range;
  // NB  does not update knob_position
  //
  Slider valueRange(float rangelow_, float rangehigh_)			// -o-
  {
    value = map(value, valueRangeLow, valueRangeHigh, rangelow_, rangehigh_);
      // XXX  nonsense upon initialization

    valueRangeLow = rangelow_;
    valueRangeHigh = rangehigh_;

    // resetting value range invalidates existing lag sequence
    knobLagActive = false;

    return this;
  }


  float value()  { return value; } 					// -o-

  // set value.  reset knob_position per value.
  //
  Slider value(float val_)
  {
    boolean  	outOfRange = false;
    int  	threshholdOfError = 1; 	
    			// XXX  ASSUME actual error is always fractional

    if (val_ < valueRangeLow) {
      if (threshholdOfError > (valueRangeLow - val_)) {
        val_ = valueRangeLow;
      } else { 
        outOfRange = true;
      }
    } else if (val_ > valueRangeHigh) {
      if (threshholdOfError > (val_ - valueRangeHigh)) {
        val_ = valueRangeHigh;
      } else { 
        outOfRange = true;
      }
    }
      
    if (outOfRange) {
      ERRORLN( className 
      		 +"(): val_="+ val_ 
		 +" out of range=["+ valueRangeLow +","+ valueRangeHigh +"]." );
    }

    value = val_;
    return resetKnobPosition();
  }  // value(float)


  // get value per knob_position.
  //
  private float valuePerPosition(int newpos_) 				// -o-
  {
    float newval;

    if (AXIS_HORIZONTAL == lengthAxis) {
      if ((newpos_ < wl ) || (newpos_ > wr)) {
        ERRORLN( className +"(): newpos_="+ newpos_
		   +" is out of range (w=["+ wl +","+ wr +"])");
      }

      newval = map(newpos_, wl, wr, valueRangeLow, valueRangeHigh);

    } else {
      if ((newpos_ < hh) || (newpos_ > hl)) {
        ERRORLN( className +"(): newpos_="+ newpos_
		   +" is out of range (h=["+ hh +","+ hl +"])");
      }

      newval = map(newpos_, hh, hl, valueRangeLow, valueRangeHigh);
    }

    return newval;
  }  // valuePerPosition()


  // set value which indirectly resets knob_position
  //
  Slider knobPosition(int percentage_) 					// -o-
  {
    if (! (percentage_ >= 0) && (percentage_ <= 100)) {
      ERRORLN( className + "(): percentage_=" + percentage_ + " out of range.");
    }

    return value( map(percentage_, 0, 100, valueRangeLow, valueRangeHigh) );
  }  


  // set knob_position per value
  //
  private Slider resetKnobPosition() 					// -o-
  {
    if (AXIS_HORIZONTAL == lengthAxis) {
      knob_position = (int) map(value, valueRangeLow, valueRangeHigh, wl, wr);
    } else {
      knob_position = (int) map(value, valueRangeLow, valueRangeHigh, hh, hl);
    }

/*
    // FIX -- debug only!  worry about slop in knob_position?
    //
    if (AXIS_HORIZONTAL == lengthAxis) {
      if ((knob_position<wl) || (knob_position>wr)) {
        WARNLN( className +"(): knob_position="+ knob_position
		  +" is out of range (w=["+ wl +","+ wr +"])");
      }
    } else {
      if ((knob_position<hh) || (knob_position>hl)) {
        WARNLN( className +"(): knob_position="+ knob_position
	  	  +" is out of range (h=["+ hh +","+ hl +"])");
      }
    }
*/

    return this;
  }  // resetKnobPosition()


  Slider knobDimentions(int w_, int h_) 				// -o-
  {
    knobWidth = w_;
    knobHeight = h_; 	
    return this;
  }


  int knobLag()  { return knob_lag; } 					// -o-

  Slider knobLag(int lag_)  
  { 
    if (0 > lag_) {
      ERRORLN( className +"(): ms lag_="+ lag_ +" must zero or greater." );
    }

    knob_lag = lag_;  

    // if knobLag is being turned off, 
    //   first move to target destination of lag in progress.
    // XXX  watch for round-off error
    //
    if ( (0 == knob_lag) && knobLagActive )
    { 
      value += (valueIncrement * valueSteps);
      valueSteps = 0;
      resetKnobPosition();

      knobLagActive = false; 
    }

    return this; 
  }  // knobLag()


  // lagStep()
  //
  //   NB  OSC triggers are sent at same rate as _fps
  //
  private boolean lagStep()  { return lagStep(-1); } 			// -o-

  private boolean lagStep(int newpos_)
  {
    float 	newval = 0;
    boolean 	gotNewPosition = false;

    if (-1 != newpos_) { 	// XXX  -1 is out of screen dimention range,
    				//        but may mask error input
      gotNewPosition = true;
      newval = valuePerPosition(newpos_);
    }
    

    // the case of no knob lag: update immediately
    //
    if (0 >= knob_lag) 
    {
      if (gotNewPosition) {
	value(newval);
	triggerAction();
	return true;
      }

      return false;
    }


    // check for and continue previously established lag step sequence.
    //
    if (! gotNewPosition)
    {
      if (! knobLagActive)  { return false; }
        // XXX  some internal changes intentionally halt current progress

      valueSequence += valueIncrement;
      valueSteps -= 1;


    // new position data: (re)start lag
    //
    } else {
      valueSteps = round( (knob_lag / 1000.0) * _fps );

      // treat knob_lag durations that are slower than the current
      //   frame rate as if lag were not active.
      //
      if (0 >= valueSteps) {
        knobLagActive = false;
        valueSequence = newval;

      } else {
	knobLagActive = true;
        valueSequence = value;

	valueIncrement = (newval - value) / valueSteps;  // may be negative

	valueSequence += valueIncrement;
	valueSteps -= 1;
      }
    }  // endif -- gotNewPosition


    if (0 >= valueSteps)  { knobLagActive = false; }

    value(valueSequence);
    triggerAction();
    return true;

  }  // lagStep()


  void triggerAction(SliderAction slidaction_) 
  {
    actionObj = slidaction_;
  }

  void triggerAction() 
  {
    actionObj.triggerAction();
  }



  boolean isActive() 							// -o--
  {
    boolean  	rval = false,
    		positionChange = false;
    int 	newpos;


    // delay prevents new input, though existing lag step sequence continues
    //
    if (isDelayed()) {
      positionChange = lagStep();
      rval = true;

    } else {
      // check for new difference data
      //
      newpos = scanSlider();

      // in the absence of difference, continue lag step sequence.
      // otherwise register new trigger and process position via lag
      //   step algorithm.
      //
      if (-1 == newpos) { 
	positionChange = lagStep();

      } else {
	positionChange = lagStep(newpos);

        if (positionChange) {
	  trigger_count += 1;
	  setDelayTime(0); 
	  rval = true;
	}
      }
    }

    if (positionChange && osc_trigger)  { oscSend(); }

    return rval;
  }  // isActive()


  // scanSlider() 							// -o-
  //
  // NB  current knob_position is never searched
  //
  private int scanSlider()
  {
    int  firsthit = -1,
         diffcount = 0;

    boolean  differenceConfirmed = false;


    scan_difflist:
    {
      if (AXIS_HORIZONTAL == lengthAxis) 	// AXIS_HORIZONTAL
      {
	if ( (knob_position-wl) < (length/2) ) 	// search right, then left
	{
	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int x = knob_position+1; x <= wr; x++) { 	// right
	    for (int y = hl; y >= hh; y--) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = x; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- y
	  }  // endfor -- x

	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int x = knob_position-1; x >= wl; x--) { 	// left
	    for (int y = hl; y >= hh; y--) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = x; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- y
	  }  // endfor -- x

	} else { 				// search left, then right
	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int x = knob_position-1; x >= wl; x--) { 	// left
	    for (int y = hl; y >= hh; y--) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = x; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- y
	  }  // endfor -- x

	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int x = knob_position+1; x <= wr; x++) { 	// right
	    for (int y = hl; y >= hh; y--) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = x; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- y
	  }  // endfor -- x
	}  // endif -- relative knob position


      } else  					// AXIS_VERTICAL
      {
	if ( (knob_position-hh) < (length/2) ) 	// search down, then up
	{
	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int y = knob_position+1; y <= hl; y++) { 	// down
	    for (int x = wl; x <= wr; x++) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = y; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- x
	  }  // endfor -- y

	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int y = knob_position-1; y >= hh; y--) { 	// up
	    for (int x = wl; x <= wr; x++) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = y; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- x
	  }  // endfor -- y

	} else { 				// search up, then down
	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int y = knob_position-1; y >= hh; y--) { 	// up
	    for (int x = wl; x <= wr; x++) 
	    {
	      diffcount += difflist[(y*width)+x];
	      if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = y; }

	      if (pixelDiffCountMinimum < diffcount) {
		differenceConfirmed = true;
		break scan_difflist;
	      }
	    }  // endfor -- x
	  }  // endfor -- y

	  firsthit = -1;
	  diffcount = 0;
	  differenceConfirmed = false;

	  for (int y = knob_position+1; y <= hl; y++) { 	// down
	    for (int x = wl; x <= wr; x++) 
	    {
	      if (1 == difflist[(y*width)+x])
	      {
		diffcount += 1;

		if ( (1 == diffcount) && (-1 == firsthit) )  { firsthit = y; }

		if (pixelDiffCountMinimum < diffcount) {
		  differenceConfirmed = true;
		  break scan_difflist;
		}
	      }
	    }  // endfor -- x
	  }  // endfor -- y
	}  // endif -- relative knob position

      }  // endif -- axis
    }  // endblock -- scan_difflist


    if (differenceConfirmed) 
    { 
      trigger_count += 1;
      if (osc_trigger) { oscSend(); }
      setDelayTime(0); 	// positive hit automatically resets delayInterval
      return firsthit;
    }

    return -1;
  }  // scanSlider()

}  // class Slider



// ------------------------------------------------ -o--
// SliderAction
//
// SliderAction.triggerAction() is invoked by Slider when Slider.value changes.
// overload triggerAction() to customize behavior of slider.
//
class SliderAction
{
  Slider slidObj;

  SliderAction(Slider slid_)  { slidObj = slid_; }

  SliderAction triggerAction()
  {
    slidObj.oscClearMsg();
    slidObj.oscMsg( round100(slidObj.value()) );

    return this;
  }
}  // class SliderAction

