/**                                                                     // -o-
 * Trigger.pde
 *
 * Trigger is the base class for Buttons and Sliders.
 * Button subclasses Rectangles and Circles as distinct from Sliders.
 *
 * Building blocks for inter-trigger logic include:
 *   . setDelayMode(), setDelayTime()
 *      + Trigger.DELAY_HOLD vs. WAIT and OPEN
 *   . isAction()
 *      + triggerCount()
 *   . oscMsg()/oscBundle()
 *      + oscPath(), oscClear*()
 *   . oscSend()
 *      + oscTrigger()
 *   . enableText()/writing()
 *
 * ----------------------------------------------------------------------
 * Copyright © 2011, David Reeder.  All rights reserved.
 * Please see the LICENSE file in this distribution for license details.
 * ----------------------------------------------------------------------
 */


// ------------------------------------------------ -o--
// Trigger implements logic for...  logging, delay, OSC, text.
//
abstract class Trigger  
{
  String  	className = "Trigger"; 

  String        osc_path;        
  OscMessage    osc_msg 	= null;  
  OscBundle 	osc_bundle 	= null;
  int 		osc_bundleSize 	= 0;
  boolean       osc_trigger 	= true; 	// enable OSC per trigger


  String        colorName;
  int           colorVal; 			// RGB color value
  int           colorFill = COLOR_CLEAR; 	// object fill option
  int 		colorFillSolid = COLOR_VISIBLE; // default "solid"

  public static final int       COLOR_SOLID = 255;  
  public static final int       COLOR_VISIBLE = 128;  // slightly translucent
  public static final int       COLOR_CLEAR = 0;


  int           delayMode       = DELAY_OPEN;
  int           delayInterval   = 0;    
  int           delayUntil      = 0; 
                        // delay until this time if delay is true,
                        //   else contains last "delayed to" time

  public static final int       DELAY_HOLD = -1;        // permanent delay
  public static final int       DELAY_OPEN = 0;         // no delay
  public static final int       DELAY_WAIT = 1;         // temporary delay


  boolean 	text_enabled 	= false;
  String 	text_string 	= "";


  int 		pixelDiffCountMinimum;
  			// local application of _pixelDiffCountRatio
			// ASSUME  area of difference blobs is
			// 	   localized and passes as a group through
			// 	   trigger space.  thinner triggers have
			// 	   less change to measure the presence of
			// 	   the whole blob.

  int           trigger_count = 0;       
                        // number of times this trigger has been enabled

  static final int  dumpTrigSpacer = 46;    
                        // maximum bytes of leading header field in trigger dump



  // -------------------------                                          -o-
  // constructors
  //
  Trigger() { /*EMPTY*/ }

  Trigger(String path_, String color_) 
  {
    osc_path     = path_; 	// XXX  differentiate with system prefix?

    colorName   = color_;
    colorVal    = Integer.parseInt( colorlist.get(colorName).toString() );
  }



  // -------------------------                                          -o-
  // common methods
  //
  abstract boolean 	   isActive();
  abstract protected void  checkBounds();


  String toString(String subclass_desc, String subclass_info) {
    return toString(subclass_desc, subclass_info, null);
  }

  String toString( String subclass_desc, 
  		   String subclass_info1, String subclass_info2 ) 
  { 
    String  delay_data 	= null;
    String  osc_data 	= null;
    String  text_data 	= null;
    String  rval 	= null;

    String  spacer_partial 	= makeSpacer(subclass_desc.length()) + "-- ";
    String  spacer_full 	= makeSpacer(-1) + "   ";


    String                                mode = "OPEN";
    if (DELAY_WAIT == delayMode)        { mode = "WAIT"; }
    else if (DELAY_HOLD == delayMode)   { mode = "HOLD"; }

    delay_data = "DELAY  "+ mode +" int="+ delayInterval 
		   +" until="+ timestamp_ms(delayUntil)
		   +"   TRIG "+ printToggle("", (true==osc_trigger)?1:0, " ")
		   +" count="+ trigger_count ;


    if (osc_trigger) {
      osc_data = "OSC  "+ dumpOsc(osc_msg, "", osc_bundle, osc_bundleSize);
    }
    if (text_enabled) {
      text_data = "   TEXT  "+ text_string;
    }


    rval = "  "+ subclass_desc + spacer_partial + subclass_info1;

    if (text_enabled)  	{ rval += text_data; } 
    if (null != subclass_info2) {
      			  rval += "\n  "+ spacer_full + subclass_info2;
    }
    			  rval += "\n  "+ spacer_full + delay_data;
    if (osc_trigger)  	{ rval += "\n  "+ spacer_full + osc_data; }


    return rval;

  }  // toString()


  Trigger display() 
  {
    strokeWeight(4); 		// XXX  make a system global value?
    stroke(colorVal);
    fill(colorVal, colorFill);

    return this;
  }



  // -------------------------                                          -o-
  // logging--
  //  
  //   XXX  className no workie
  //
  private void logln(String loglevel, String s) {
    println( 
      loglevel +"  "+ className +"("+ osc_path +", "+ colorName + ") :: " + s );
  }

  void ERRORLN(String s)  { this.logln("ERROR", s); exit(); }
  void WARNLN(String s)   { this.logln("WARNING", s); }
  void DEBUGLN(String s)  { if (1 == debug) { this.logln("DEBUG", s); } }
  void NOTELN(String s)   { this.logln("NOTE", s); }

  void ASSERTLN(String s)  { this.logln("ASSERT", s); exit(); }
  void ASSERT(boolean valid, String s)  { if (!valid) { ASSERTLN(s); } }



  // -------------------------                                          -o-
  // delay logic--
  //
  //   Trigger  setDelayMode(int mode_or_ms_)  [ALIAS: delay]
  //   Trigger  setDelayTime(int ms_) 
  //   boolean  isDelayed()
  //   int      triggerCount()  
  //

  // (..., 0] --> DELAY_HOLD, DELAY_OPEN
  // [1, ...) --> DELAY_OPEN and store delayInterval
  //
  //   XXX  use enums (here and elsewhere)
  //   NB  DELAY_HOLD is set (and reset) only by setDelayMode()
  //
  Trigger delay(int mode_or_ms_)  { return setDelayMode(mode_or_ms_); }

  Trigger setDelayMode(int mode_or_ms_) 
  { 
    delayInterval = 0;
    delayMode = DELAY_OPEN;

    if (0 < mode_or_ms_) {
      delayInterval = mode_or_ms_; 

    } else {
      delayMode = min(DELAY_HOLD, mode_or_ms_);
    }

    return this;
  }


  // use delayInterval if ms_ <= 0; 
  // change delayMode if delayUntil is updated
  //
  Trigger setDelayTime(int ms_) 
  { 
    int interval = (0 >= ms_)? delayInterval: ms_; 

    if (0 < interval) {
      delayMode = DELAY_WAIT;
      delayUntil = millis() + interval;
    }
    
    return this;
  }


  // colorFill setting reflects delayMode;
  // resets DELAY_WAIT -> DELAY_OPEN when delayUntil expires
  // 
  // CALLED BY: isActive()
  //
  protected boolean isDelayed()
  {
    boolean rval = true;  // ASSUME: DELAY_WAIT && (delayUntil > millis())

    if (DELAY_HOLD == delayMode) {
      rval = true;

    } else if ( (DELAY_OPEN == delayMode) || (millis() > delayUntil) ) {
      delayMode = DELAY_OPEN;
      rval = false;
    }

    colorFill = rval? colorFillSolid: COLOR_CLEAR;
    return rval;        

  }  // isDelayed()


  int triggerCount()  { return trigger_count; }



  // -------------------------                                          -o-
  // OSC message logic--
  //
  //   void     oscTrigger(boolean onoff_) 
  //   boolean  oscTrigger()         
  //
  //   void     oscClearMsg()              
  //   void     oscClearBundle()
  //
  //   void     oscPath(String path_)
  //   void     oscMsg(float ... args_)        
  //   void     oscMsg(String ... args_)        
  //   void     oscBundle()
  //
  //   void     oscSend()       
  //
  // USAGE
  //   . enable/disable osc for a given trigger 
  //   . add arguments to a message; or, clear its arguments
  //   . build a bundle one message at a time; or, clear the bundle
  //       + adding message to bundle automatically clears message field
  //   . reset the path for a message
  //   . send bundle if it exists, else send message
  //
  Trigger oscTrigger(boolean onoff_)	{ osc_trigger = onoff_; return this; }
  boolean oscTrigger()               	{ return osc_trigger; }

  Trigger oscClearMsg() {  						// -o-
    if (null != osc_msg)  { osc_msg.clearArguments(); } 
    return this;
  }

  Trigger oscClearBundle() {  
    if (null != osc_bundle)  { osc_bundle.clear(); } 

    osc_bundle 	    = null;
    osc_bundleSize  = 0;

    return this;
  }

  // NB  allows message path to diverge from that of the enclosing
  //     trigger object
  //
  Trigger oscPath() {  							// -o-
    if (null == osc_msg)  { osc_msg = new OscMessage(osc_path); }
    osc_msg.setAddrPattern(osc_path);
    return this;
  }

  Trigger oscPath(String path_) { 
    if (null == osc_msg)  { osc_msg = new OscMessage(osc_path); }
    osc_msg.setAddrPattern(path_);
    return this;
  }

  // XXX  all numeric values sent as floats...
  //
  Trigger oscMsg(float ... args_)  					// -o-
  {
    if (null == osc_msg)  { osc_msg = new OscMessage(osc_path); }
    for (int i = 0; i < args_.length; i++)  { osc_msg.add(args_[i]); }
    return this;
  }

  Trigger oscMsg(String ... args_)         
  {
    if (null == osc_msg)  { osc_msg = new OscMessage(osc_path); }

    float num = 0;
    boolean stringIsNum;

    for (int i = 0; i < args_.length; i++) {
      stringIsNum = true;

      try {
        num = new Float(args_[i]);
      } catch (Exception e) {
        stringIsNum = false;
      }

      if (true == stringIsNum) {
        osc_msg.add(num);
      } else {
        osc_msg.add(args_[i]); 
      }
    }

    return this;
  }  // oscMsg(<array of strings>)


  Trigger oscBundle()  							// -o-
  {
    if (null == osc_msg)     { this.ERRORLN("oscBundle(): osc_msg is null."); }
    if (null == osc_bundle)  { osc_bundle = new OscBundle(); }

    osc_bundle.add(osc_msg);
    osc_bundleSize += 1;

    osc_bundle.setTimetag(osc_bundle.now()); 	// XXX  need time tags?

    osc_msg.clear();
    osc_msg = null;
    
    return this;
  }  // oscBundle()


  // send bundle if it exists, otherwise send message
  //
  Trigger oscSend()        						// -o-
  {
    if (0 == enableOsc) { return this; }

    if (null != osc_bundle) 	{ osc.send(osc_bundle, osc_target); } 
    else if (null != osc_msg) 	{ osc.send(osc_msg, osc_target); } 
    else { 
      this.ERRORLN("oscSend(): osc_msg and osc_bundle are null."); 
    }


    if (1 == debug) { 
      DEBUGLN( "OSC "+ dumpOsc(osc_msg, osc_path, osc_bundle, osc_bundleSize));
    }

    return this;
  }  



  // -------------------------                                          -o-
  // text logic--
  //
  Trigger enableText(boolean onoff_)	{ text_enabled = onoff_; return this; }
  boolean enableText()               	{ return text_enabled; }

  Trigger writing() 			{ text_string = ""; 	return this; }
  Trigger writing(String text_) 	{ text_string = text_; 	return this; }
  Trigger writing(int num_) { 
    text_string = Integer.toString(num_); 			return this; }
  Trigger writing(float num_) { 
    text_string = Float.toString(num_); 			return this; }

}  // class Trigger




abstract class Button extends Trigger                           	// -o-
{
  String className = "Button";

  int 	textHOffset = 8; 	// XXX  properly center button text

  Button() { /*EMPTY*/ }
  Button(String path_, String color_) { super(path_, color_); }

}  // class Button

