/** 									// -o-
 * dansemuse.pde
 *
 * Interface to PDE.  Global system variables.
 *
 * ----------------------------------------------------------------------
 * Copyright © 2011, David Reeder.  All rights reserved.
 * Please see the LICENSE file in this distribution for license details.
 * ----------------------------------------------------------------------
 */
 
import processing.video.*;

import oscP5.*;
import netP5.*;



// ------------------------------------------------ -o--
// globals 	
//
// See overlay.pde for declaration and definition of and tests on triggers.
//
String[] camDevices;
int camDevice_index = 2; 	// XXX  convert to heuristic and/or user choice

Capture cam;
int[] 	cam_prev;
int[] 	difflist; 		// XXX  reduce to bitfield?

MovieMaker movie;


OscP5 		osc;
NetAddress 	osc_target;


HashMap 	colorlist; 	// XXX  localize to Trigger object?
boolean 	useDrawCycle = false; 	
				// true after _initialFrameTossCount frames



// ------------------------------------------------ -o-
// system config
// 
String  _version = "0.7.1";

int   _fps 	= 15,		// frames per second
      _width 	= 800, 	 	// width of input image
      _height 	= 600, 		// height of input image

      _colorDiffMinimum = 55, 	
	  // minimum difference between consecutive R, G, or B values
	  //   necessary to signify a difference at that pixel
	  // XXX  different minimum for each of R, G, B?

      _pixelDiffCountRatio = 5,
      	  // governs minimum number of pixels within the shape of a 
	  //   given trigger that must be different in order to enable 
	  //   the trigger.  
	  // formula: min(length, width) / ratio
	  // XXX  need this be object specific?

      _initialFrameTossCount = 1;
          // number of initial frames to ignore.  thus triggers (and
	  //   osc) are not activated until camera image is stable...
	  // threshhold is used to set useDrawCycle.

String  _sessionTimestamp = timestamp();


	  // paramters of saveFrame() files and filenames
	  //
String 	_displayCaptureDir 	= "capture/frame/",
	_displayCaptureName 	= "dm",
 	_displayCaptureNumber 	= "-####",
	_displayCaptureSuffix 	= ".png",
	_displayCaptureFilename = _displayCaptureDir + _displayCaptureName 
	  			    +"-"+ _sessionTimestamp 
				    +"-"+ _displayCaptureNumber 
				    + _displayCaptureSuffix;

	  // parameters of MovieMaker files and filenames
	  //
int 	_movieNumber 	= 0, 			// initial value
	_movieFPS 	= _fps,
 	_movieCodec 	= MovieMaker.JPEG,
	_movieQuality 	= MovieMaker.LOSSLESS;

String 	_movieDir 	= "capture/movie/",
 	_movieName 	= "dm",
	_movieSuffix 	= ".mov",
	_movieFilename 	= _movieDir + _movieName 
	                    +"-"+ _sessionTimestamp
	                    +"-"+ Integer.toString(_movieNumber) + _movieSuffix;


// runtime options
//
int  captureVideo = 1, 		// stop/start draw() cycle
     debug = 0, 		// enable/disable debug 
     displayCapture = 0, 	// capture display at next draw()
     enableOsc = 1, 		// disable/enable OSC
     enableDumpRGB = 0, 	// enable/disable RGB color test
     movieRecord = 0, 		// enable/disable video documentation
     showDifferance = 1, 	// show/don't difference between frames
     showOverlay = 1, 		// disable/show overlay of triggers
     showRawInput = 0, 		// show/don't raw camera input
     visualFeedback = 1; 	// disable/enable visual feedback



// ------------------------------------------------ -o--
void setup() 	
{
  // interface to camera
  //
  size(_width, _height, P2D); 	// XXX  can these be reset dynamically?
  frameRate(_fps);	

  noCursor(); 	// XXX  optimize for speed...
  noSmooth();


  splash(); 	// XXX  any sooner and it prints twice (?)


  // camera device; current and previous frames; 
  //   array to summarize differences between current and previous frames
  //
  camDevices = Capture.list();

  cam 	   = new Capture( this, width, height, 
  		            camDevices[camDevice_index], _fps );		
			    	// XXX  can these be reset dynamically?
  cam_prev = new int[cam.width * cam.height];

  difflist = new int[cam.width * cam.height];


  // OSC configuration lives in overlay module
  //
  setupOsc();


  // establish system-wide font
  //
  PFont font = loadFont("HelveticaNeue-28.vlw"); 
  textFont(font, 28); 
  textMode(SCREEN); 		// XXX  optimize per size(..., P2D)
  textAlign(CENTER); 


  // setup and define triggers; dump system values
  //
  setupTriggers();
  defineTriggers();

  dumpSystemValues();

} // setup()




// ------------------------------------------------ -o--
void draw() 
{
  int  	   index, cam_index, pix, pix_prev; 	// traverse image
  int  	   dr, dg, db; 				// color differences

  boolean  differenceDetected; 			// register difference


  if (cam.available()) 
  {
    // prep for traversal
    //
    cam.read(); 
    cam.loadPixels(); 	
    cam_index = 0;
    
    if (1 == visualFeedback) {
      loadPixels();
    }


    // traverse input image
    //
    // note differences, detect localized change, update control feedback.
    // NB: traverse width backwards
    //
    for (int h=0; h < cam.height; h++) 
    {
      for (int wr=(width-1); wr >= 0; wr--) 	// width reversed
      {
	pix 	 = cam.pixels[cam_index];
	pix_prev = cam_prev[cam_index];

	differenceDetected = false;
	index = (h*width)+wr; 			// mirror image


	// optionally dump data on RGB color change differences 
	// XXX  toss in production releases
	//
	if (1 == debug) {
	  dumpRGB(wr, h, pix, pix_prev);
  	}


	// test for change
	//
	dr = abs((pix & 0x00ff0000)-(pix_prev & 0x00ff0000)) >> 16;
	dg = abs((pix & 0x0000ff00)-(pix_prev & 0x0000ff00)) >> 8;
	db = abs((pix & 0x000000ff)-(pix_prev & 0x000000ff));
     
	if ( (_colorDiffMinimum < dr) 
	       || (_colorDiffMinimum < dg) || (_colorDiffMinimum < db) )
	{
	  differenceDetected = true;
	}

        // capture and display differences
	//
	if ((1 == visualFeedback) && (1 == showDifferance)) {
	  pixels[index] = 0; 			// clear to black
	}
	difflist[index] = 0;

	if (differenceDetected) {
	  if ((1 == visualFeedback) && (1 == showDifferance)) {
	    pixels[index] = 0x00FFFFFF; 	// difference is white
	  }
	  difflist[index] = 1;
	}


        // optionally display camera input
	//
	// XXX  presumably inline is preferable to "image(cam,0,0)"
	//      because feedback suffers similar latency to
	//      difference computation
	//
	if ((1 == visualFeedback) && (1 == showRawInput)) {
	  pixels[index] = pix; 
	}


	cam_prev[cam_index] = pix;
	cam_index += 1;

      }  // endfor -- wr
    }  // endfor -- h
  }  // cam.available()



  // image bookkeeping and capture
  //
  if (useDrawCycle) { 	
    scanTriggers();

  } else if (_initialFrameTossCount < frameCount) {
    useDrawCycle = true;
    	// NB  at _fps=15, frameCount wraps after nine years...
  }

  if (1 == showOverlay) { displayTriggers(); }

  if (1 == visualFeedback) { updatePixels(); }

  if (1 == displayCapture) { 			// capture screenshot
    displayCapture = 0;
    saveFrame(_displayCaptureFilename);
  }

  if (1 == movieRecord)  { movie.addFrame(); } 	// real-time display capture

}  // draw()



// slash() 								// -o-
//
// Vanity function.
//
// FIX  adapt to LICENCE, ers se.
// TBD  graphic
//
void splash()
{
  String[] splashtext = {
    "",
    "dansemuse -- v" + _version,
    "",
    "(C)opyright 2011 by David Reeder",
    "dansemuse@danaprajna.com",
    "",
    "Type 'h' for help.",
    "",
    "",
  };

  for (int i = 0; i < splashtext.length; i++)  { println(splashtext[i]); }

  delay(5000);

}  // splash()

